import torch
from typing import Dict

# Local imports
from .utils.litgpt_minimal.model import GPT
from .utils.litgpt_minimal.tokenizer import Tokenizer
from .utils.utils import clean_token
from .llm import LLM

class Patcher(LLM):
    def __init__(self, model:GPT, tokenizer:Tokenizer, prompt: str, placeholder_token:str):
        self.model = model
        self.tokenizer = tokenizer
        self.prompt = prompt
        self.placeholder_token = placeholder_token

        # Initialize the device
        self._init_device()
        
        # Set the default generated parameters
        self.generated_params = self._generated_params(top_k = 50, temperature = 0.0, max_returned_tokens = 100)
    
    def merged_hs(self, hidden_states: torch.Tensor, weights: torch.Tensor, verbose = True) -> torch.Tensor:

        # Merge the hidden states of each hidden layer
        merged_hs = list()
        for i in range(hidden_states.shape[0]):
            
            # Merge the hidden states using a weighted sum
            ws = torch.sum(hidden_states[i] * weights, dim = 0)
            merged_hs.append(ws)
            
            if verbose:
                print(f'L{i}:', hidden_states[i].shape, '-->', ws.shape)
        merged_hs = torch.stack(merged_hs)

        return merged_hs
    
    def _single_patching(self, hidden_states: torch.Tensor):

        # Tokenize the target prompt
        input_ids, input_tokens = self._tokenize(self.prompt, return_tokens = True)
        
        # Retrive the position of the placeholder token
        patching_token_pos = [idk for idk, token in enumerate(input_tokens) if clean_token(token) == self.placeholder_token]

        if len(patching_token_pos) != 1:
            raise ValueError(f"The placeholder token '{self.placeholder_token}' should appear once in the prompt! TOKENS:{input_tokens}")
        patching_token_pos = patching_token_pos[0]

        assert clean_token(input_tokens[patching_token_pos]) == self.placeholder_token, f"The placeholder token '{self.placeholder_token}' should appear at the position {patching_token_pos} in the prompt! FOUND INSTEAD: {clean_token(input_tokens[patching_token_pos])}"
        
        # Inject the hidden states into the model
        self.model.load_source_hidden_states(hidden_states, patching_token_pos = patching_token_pos)
        
        # Perform inference patched with the hidden states
        generated_text, _ = self._generate(input_ids, parse_output=True)

        return generated_text

    def probing(self, hidden_states: torch.Tensor) -> Dict[str, str]:
        
        generated_texts = dict()
        
        # Apply patching for each hidden layer
        num_layers = hidden_states.shape[0]
        
        # TODO CAN BE PARALLELIZED 
        for i in range(num_layers):
            generated_texts[f'L{i}'] = self._single_patching(hidden_states[i])

        return generated_texts



from os import path, makedirs
from json import dump
from typing import List, Dict

def save_results(results: List[Dict[str, str]], folder_path:str, file_name:str = 'results.json'):
    
    # Create the folder if it does not exist
    if not path.exists(folder_path):
        makedirs(folder_path)
    
    file_path = path.join(folder_path, file_name)
    with open(file_path, mode = 'w', encoding='utf-8') as f:
        dump(results, f, indent = 4, ensure_ascii = False)
    
    print(f"\nThe results are saved in '{file_path}'!\n")


from collections import defaultdict
from logging import warn
from pathlib import Path
from multiprocessing.pool import ThreadPool 
import sys
from typing import Union, Tuple, List, Dict
from lightning import Fabric
import torch
#import warnings

# LITGPT IMPORTS
from litgpt import Config
from litgpt.scripts.download import download_from_hub
from litgpt.utils import check_valid_checkpoint_dir, get_default_supported_precision
from litgpt.generate.sequentially import sequential

# LOCAL IMPORTS
from .utils.litgpt_minimal.model import GPT
from .utils.litgpt_minimal.base import generate
from .utils.litgpt_minimal.tokenizer import Tokenizer
from .utils.utils import print_title, text2json, generate_prompt_template, search_subvector, clean_token
from .utils.data_loader import load_examples

class LLM:
    
    DEFAULT_SOURCE_SYS_INST = "You are a journalist with expertise in fact-checking. Your role is to evaluate the truthfulness of factual claims. To uphold journalistic integrity, you must produce a report containing a binary assessment and all the factual information that supports your evaluation."
    
    DEFAULT_TARGET_SYS_INST = "You are an assistant with expertise in fact-checking. Your role is to assess claims."
    
    ICL_DEFAULT_EXAMPLES_PATH = Path('src/latent_explorer/resources/in_context_examples.json')
    
    def __init__(self, model_name:str, inputs: list[str], hf_access_token:str = None):
        self.model_name = model_name
        self.inputs = inputs
        
        # Initialize the device
        self._init_device()

        # Access token for downloading the model
        self.hf_access_token = hf_access_token
        
        # Set the placeholder token
        self.placeholder_token = "x"
        
        # Visualize the model information
        print_title(self.__str__())
        
        # Load the model
        self._download_model()
        self._load_model()
        
    def __str__(self) -> str:
        return f'LatentExplorer: MODEL={self.model_name}, DATA={len(self.inputs)}, DEVICE={self.fabric.device.type.upper()}(x{self.num_devices})'
    
    def get_model_name(self) -> str:
        return list(reversed(self.model.config.hf_config.values()))

    def _init_device(self):
        precision = get_default_supported_precision(training=False)
        
        self.fabric = Fabric(devices = 1, precision=precision)
        self.num_devices = self.fabric.accelerator.auto_device_count()
        
    def _download_model(self):
       
        # Create the model folder if it does not exist  
        model_folder = Path('.models')
        model_folder.mkdir(parents=True, exist_ok=True)
        
        # Download the model if not exists
        self.model_path = model_folder.joinpath(self.model_name)
        if not self.model_path.exists():
            print(f"\n[INFO] Downloading the model from Hugging Face: {self.model_name}\n")
            
            try:
                download_from_hub(repo_id = self.model_name, access_token = self.hf_access_token, checkpoint_dir = model_folder) 
            except ValueError as e:
                if self.hf_access_token is None:
                    print(f'[ERROR] {self.model_name} requires authentication, please set the `hf_access_token` variable in the LatentExplorer class. You can find your token by visiting https://huggingface.co/settings/tokens')
                else:
                    print(f'[ERROR] {e}')
                sys.exit(1)
            print(f'\n[INFO] Model downloaded: {self.model_name}')
    
    def _load_model(self):
        
        # Load the metadata
        check_valid_checkpoint_dir(self.model_path)
        config = Config.from_file(self.model_path / "model_config.yaml")
        checkpoint_path = self.model_path / "lit_model.pth"

        # Load the tokenizer
        self.tokenizer = Tokenizer(self.model_path)

        # Initalize the model
        with self.fabric.init_tensor(), torch.device("meta"):
            model = GPT(config)
            model.set_kv_cache(batch_size = 1)

        # Load the model checkpoints
        state_dict = torch.load(str(checkpoint_path), mmap=True, map_location="cpu", weights_only=False)
        model.load_state_dict(state_dict, assign=True)

        # Setup the model
        model = self.fabric.setup_module(model, move_to_device=False)

        # TURNAROUND for NotImplementedError: Only balanced partitioning is implemented
        while model.config.n_layer % self.num_devices != 0:
            self.num_devices-= 1
            
        # Partition the transformer blocks across all your devices and running them sequentially (https://github.com/Lightning-AI/litgpt/blob/main/tutorials/inference.md)
        print(f'Loading the model "{self.model_name}" on {self.num_devices} {self.fabric.device.type.upper()} devices:')
        self.model = sequential(model, root = self.fabric.device, max_seq_length = config.block_size, devices = self.num_devices)
        
        if self.fabric.device.type == "cuda":
            self.fabric.print(f"\nCUDA memory used: {torch.cuda.max_memory_allocated() / 1e9:.02f} GB")
            
        # Set the default generated parameters
        self.generated_params = self._generated_params(top_k = 50, top_p = 0.9, temperature = 0.0, max_returned_tokens = 512)
    
    def _generated_params(self, **kwargs):
        params = dict(eos_id = self.tokenizer.eos_id)
        if len(kwargs.keys()) > 0:
            params.update(kwargs)

        return params
    
    def _tokenize(self, prompt:str, return_tokens:bool) -> torch.Tensor:
        
        # Input encoding
        input_ids = self.tokenizer.encode(prompt, device = self.fabric.device, add_special_tokens = False)
        
        if return_tokens:
            tokens = [self.tokenizer.id_to_token(token.item()) for token in input_ids]
            return input_ids, tokens
        
        return input_ids
    
    def _generate(self, input_ids:torch.Tensor, parse_output:bool) -> Tuple[str, torch.Tensor]:
        
        # Set the generated parameters
        params = self.generated_params.copy()
        params['max_returned_tokens'] += input_ids.size(0)
        
        if params['max_returned_tokens'] > self.model.config.block_size:
            warn(f" Setting 'max_returned_tokens' to {self.model.config.block_size} since it is greater than the block size: {params['max_returned_tokens']} > {self.model.config.block_size}\n")
            params['max_returned_tokens'] = self.model.config.block_size
        
        # Generate the output
        output_ids, hidden_states = generate(model = self.model, prompt = input_ids, output_hidden_states = True, **params)
        
        # Decode the output
        output_ids = output_ids[input_ids.size(0):]
        generated_text = self.tokenizer.decode(output_ids).strip()
        
        # Parse the output text to JSON
        if parse_output and len(generated_text) > 0:
            generated_text = text2json(generated_text, max_attempts = 3)
        
        return generated_text, hidden_states
    
    def _inference(self, input_ids:torch.Tensor, parse_output:bool = True, output_hidden_states:bool = True) -> Union[str, Tuple[str, torch.Tensor]]:
        
        # Generate the output
        generated_text, hidden_states = self._generate(input_ids, parse_output = parse_output)
        
        if output_hidden_states:
            return generated_text, hidden_states
        else:
            return generated_text
    
    # --------------------------------------------
    # ---------- PUBLIC METHODS ------------------ 
    # --------------------------------------------
    def generate_prompts(self, source_sys_inst:str = None, target_sys_inst:str = None, in_context_examples:list[dict] = None, verbose = False):
        
        # Set the system instructions
        source_sys_inst = source_sys_inst if source_sys_inst is not None else self.DEFAULT_SOURCE_SYS_INST
        target_sys_inst = target_sys_inst if target_sys_inst is not None else self.DEFAULT_TARGET_SYS_INST
        models_w_sys_roles = ['llama']
        
        # Load the default in-context examples
        if not isinstance(in_context_examples, list) or len(in_context_examples) < 1:
            in_context_examples = load_examples(self.ICL_DEFAULT_EXAMPLES_PATH)
    
        # Split the examples into source and target examples (1st example for the source, the others for the target)
        if in_context_examples == 1:
            source_examples = target_examples = in_context_examples
        else:
            source_examples = [in_context_examples[0]]
            target_examples = in_context_examples[1:]
        
        # Generate the source prompts for the inputs using different threads
        generate_prompt = lambda input: self.tokenizer.apply_chat_template(
            conversation = generate_prompt_template(
                sys_prompt = source_sys_inst, 
                examples = source_examples, 
                input_text = input, 
                role_sys_available = any([m in self.model_name.lower() for m in models_w_sys_roles])), 
            tokenize = False)
        
        with ThreadPool(processes = len(self.inputs)) as pool:
            result = pool.map_async(func = generate_prompt, iterable = self.inputs)
            self.prompts = [value for value in result.get()]
            
        # Generate the target prompts for the inputs using different threads
        self.target_prompt = self.tokenizer.apply_chat_template(
            conversation = generate_prompt_template(
                sys_prompt = target_sys_inst, 
                examples = target_examples, 
                input_text = self.placeholder_token, 
                role_sys_available = any([m in self.model_name.lower() for m in models_w_sys_roles])), 
            tokenize = False)
        
        # Save the tokenized input texts (for part-of-speech tagging)
        self.input_tokens = [self._tokenize(input, return_tokens=True)[1] for input in self.inputs] 
        
        if verbose:
            print_title('SOURCE PROMPT TEMPLATE')
            print(self.prompts[0].replace(self.inputs[0], "$INPUT"), end = '\n' + '-' * 100 + '\n\n')
            print_title('TARGET PROMPT')
            print(self.target_prompt, end = '\n' + '-' * 100 + '\n\n')
        
    def inference(self, parse_output:bool = True, output_hidden_states:bool = True, verbose:bool = False)-> List[Dict[str, Union[str, torch.Tensor]]]:

        # Perform inference for each prompt
        self.outputs = list()
        for idk, prompt in enumerate(self.prompts):
            
            # Tokenize the input text
            input_ids, input_tokens = self._tokenize(prompt, return_tokens=True)
            
            # Find the input tokens within the prompt
            pos = search_subvector(vector = input_tokens, sub = self.input_tokens[idk])
            self.input_tokens[idk] = dict(zip(pos, self.input_tokens[idk]))        
            
            # Perform inference
            generated_text, hidden_states = self._inference(
                input_ids = input_ids, 
                parse_output = parse_output, 
                output_hidden_states = output_hidden_states
                )
            
            # Save the output as a dictionary
            self.outputs.append({'text': generated_text, 'hs': hidden_states})
            
            if verbose:
                print_title('PROMPT')
                print(prompt)
                print_title(f'OUTPUT [hs: {tuple(hidden_states.size())}]')
                print(generated_text, end = '\n\n')

        return self.outputs


from json import load
from litgpt.config import configs

def load_examples(path: str) -> list[dict]:
    with open(path, mode = 'r', encoding="utf-8") as f:
        examples = load(f)
    return examples

def all_supported_models(verbose = True) -> list[str]:
    supported_models = [f"{config['hf_config']['org']}/{config['hf_config']['name']}" for config in configs]
    supported_models = list(sorted(supported_models))
    
    if verbose:
        print("\n".join(f'[{i}] {model}'for i, model in enumerate(supported_models)))
        
    return supported_models


from json import dumps
from re import split
import dirtyjson
import numpy as np
import pandas as pd

def generate_prompt_template(sys_prompt = None, examples: list[dict] = [], input_text:str = '', role_sys_available = False) -> str:
    conversation = []
    
    # Add the system prompt
    if role_sys_available and sys_prompt is not None:
        conversation.append({"role": 'system', 'content': sys_prompt})
    else:
        conversation.append({"role": 'user', 'content': sys_prompt})
    
    # Add the in-context examples
    for pair in examples:
        claim, output = list(pair.items())[0]
        
        if not role_sys_available and len(conversation) == 1: # User input: Add the claim to the system instruction if the system role is not available
            conversation[0]['content'] += f"\n\n{claim}"
        else:  # User input: Add just the example claim to conversation if the system role is available
            conversation.append({"role": 'user', 'content': claim})
        
        # Assistant output
        conversation.append({"role": 'assistant', 'content': dumps(output, ensure_ascii = False)})
    
    # Add the input text
    if input_text:
        conversation.append({"role": 'user', 'content': input_text})

    return conversation

def text2json(output, max_attempts = 3, verbose = True):
    
    # Attempt to parse the output as JSON with some error handling
    num_attempts = 0
    while num_attempts < max_attempts:
        
        # Attempt to parse the output as JSON
        try:
            return dirtyjson.loads(output)
        
        # Handle the parsing error
        except dirtyjson.error.Error as e:
            
            # Return the output if the text is empty
            if len(e.doc) == 0:
                return output
            
            # Visualize the warning and error
            if verbose: 
                print('\t' * num_attempts + f'[WARNING] Parsing Error ({num_attempts + 1}° attempt):', e, '-->', 
                      f'[{e.doc[e.pos].upper()}]{e.doc[e.pos + 1:][:5]}...' if e.pos + 5 < len(e.doc) else f'[{e.doc[-1]}]')
                
                # Last attempt: Print the final output
                if num_attempts == (max_attempts - 1):
                    print('\t' * (num_attempts + 1), '-->', output, "\n")
            
            # FIX A: Error at the last character            
            if e.pos < len(e.doc): 
                if e.doc[e.pos] == '"' or e.doc[e.pos - 1] == '"':
                    output = output[:e.pos] + "]}"
            else:
                output += "]}" 
                
            # FIX B: a fully unstructured string
            if e.pos <= 1:
                splitted_output = [item.strip() for item in output.split('\n') if item.strip()]
                print('RAW [SPLITTING]:', splitted_output)
                
                if len(splitted_output) > 1:
                    output = splitted_output[1] # skip the first component (e.g., assistant:\n ...)
                else:
                    return output.strip()
                    
            # Increment the attempt counter
            num_attempts += 1
        
        # Handle the general exception
        except Exception as e:
            print('[WARNING] Parsing General Error:', e)
            return output
    return output

def print_title(title, char = "-", length = 150):
    pad = length - len(title) - 2
    l_pad = pad // 2
    r_pad = (pad // 2) + (pad % 2)
    
    print("\n" + char*length)
    print(char * l_pad, title, char * r_pad)
    print(char*length + "\n")
    
def clean_token(token) -> str:        
    if len(token) <= 1:
        return token
    
    # Remove the special characters added by the tokenizer (artifacts of the tokenization process)
    special_chars = ['▁', 'Ġ']

    return token.strip(''.join(special_chars))
    
def search_subvector(vector, sub):
    
    # Clean the token
    vector = [clean_token(token) for token in vector]
    sub = [clean_token(token) for token in sub]
    
    # Reshape the vector and sub-vector
    vector = np.array(vector).reshape(1, -1)
    sub = np.array(sub).reshape(-1, 1)
    
    # Element-wise comparison
    comparison = np.argwhere(vector == sub)

    # Convert the comparison to a DataFrame and group by the first dimension
    df = pd.DataFrame(comparison)

    # Group by the first dimension and merge the second dimension
    df = df.groupby(by = 0)[1].apply(list)
    df = df.map(np.max).to_list()
    
    # Check if entire sub-vector is found in the vector, otherwise raise an error
    if len(df) == len(sub):
        return df
    else:
        print(f'\nORIGINAL: {sub[:, 0]}\nFOUND: {dict(zip(df, vector[0, df[0]:]))}')
        raise LookupError("The sub-vector is not found in the vector.")


from os import path

from latent_explorer import LatentExplorer, TempoGrapher
from latent_explorer.utils import save_results, all_supported_models

data = [
    "Empress Matilda moved to Germany as a child",
    "Robin was murdered by the Joker in a 1989 book",
    "Bojack Horseman's creator is also American"
]

if __name__ == '__main__':
    
    # List all supported models
    models = all_supported_models(verbose = True) 

    # APP INITALIZATION: Initialize the application
    explorer = LatentExplorer(model_name = "meta-llama/Llama-2-7b-chat-hf", inputs = data, hf_access_token = None) 

    # APP INITALIZATION: Generate the textual prompts (encompassing the system prompt, examples, and input text)
    explorer.generate_prompts(verbose = True)
    
    # (1) Basic inference (generated text, hidden states)
    explorer.inference(parse_output = True, output_hidden_states = True, verbose = False)

    # (2) Patched inferences (for each hidden states)
    results = explorer.probe_hidden_states(verbose = False)

    # (3) Save the results
    output_folder = path.join("outputs", *explorer.get_model_name())

    # (3a) Save the statistics
    explorer.save_stats(folder_path = output_folder)
     
    # (3b) Save the results as a JSON file
    save_results(results, folder_path = output_folder)
    
    # (4) Generate the temporal knowledge graphs
    tg = TempoGrapher(results)
    graphs = tg.get_graphs()

    # (5) Save the graphs
    tg.save_graphs(folder_path = output_folder)


from typing import List, Dict
import torch
from collections import defaultdict
from tqdm import tqdm 
import spacy
from json import dump 
from os import path, makedirs

# LOCAL IMPORTS
from .llm import LLM
from .patcher import Patcher
from .utils.utils import clean_token

class LatentExplorer(LLM):
    def __init__(self, model_name:str, inputs: list[str], hf_access_token:str = None):
        super().__init__(model_name, inputs, hf_access_token)
        
        # Perform part-of-speech tagging
        self._pos_tagging()
        
        # Weigh the input tokens based on their part-of-speech tags
        self.tag_weights = {'NOUN': 1, 'PROPN': 1, 'VERB': 1}
        
    def _pos_tagging(self):
    
        # Check if the spacy model is installed
        if not spacy.util.is_package("en_core_web_trf"):
            spacy.cli.download("en_core_web_trf")
        
        # Load the spacy model
        spacy.prefer_gpu()
        nlp = spacy.load("en_core_web_trf", exclude=["lemmatizer", "ner"])
        
        # Perform part-of-speech tagging
        self.pos_tags = list()
        for item in nlp.pipe(self.inputs):
            
            # Extract the part-of-speech tags for each input
            item_pos_tags = defaultdict(list)
            for token in item:
                item_pos_tags[token.pos_].append(token.text)
            self.pos_tags.append(dict(item_pos_tags))
    
    def _weigh_input_tokens(self, hidden_states:torch.Tensor, verbose:bool = False) -> List[torch.Tensor]:
               
        item_weights = []
        item_considered_tokens = []
        for i in range(len(self.inputs)):
            
            # Get the input text
            input_text = self.inputs[i]
            
            # Get the part-of-speech tags for the input tokens
            pos_tags = self.pos_tags[i]
            
            # Filter the eligible words
            word_weights = {word: self.tag_weights[pos_tag] 
                            for pos_tag, words in pos_tags.items() 
                            for word in words 
                            if pos_tag in self.tag_weights.keys()}
            
            # EXPERIMENT exp1: CONSIDER ALL THE WORDS
            #word_weights = {word: 1 for words in pos_tags.values() for word in words}
            
            # Initialize the weights to zeros
            num_tokens = hidden_states[i].shape[1]
            weights = torch.zeros(num_tokens)
            
            # Get the input tokens (pos: token)
            input_tokens = self.input_tokens[i]
            
            # Compute the weights
            for pos, token in input_tokens.items():
                
                # Clean the token by removing the special characters
                token = clean_token(token)
                
                # Skip the token if it only contains one character
                if len(token) < 2: # exp1
                    continue
                
                # Check if the token includes the the end of the considered words
                for word, word_weight in word_weights.items():
                    if word.endswith(token):
                    #if token in word: # exp1 & exp2
                        weights[pos] = word_weight

            # The weights before the first considered token should be zeros
            assert  torch.argwhere(weights > 0).flatten()[0] >= list(input_tokens.keys())[0], f"The token weights before the input text ({torch.argwhere(weights > 0).flatten()[0]} < {list(input_tokens.keys())[0]}) should be zeros!"

            # Save the considered tokens
            considered_tokens = {token: weights[pos].item() for pos, token in input_tokens.items()}

            # Save the considered tokens and their weights 
            item_considered_tokens.append(considered_tokens)
            item_weights.append(weights)
            
            if verbose:
                print("\nINPUT:", input_text)
                print("POS TAGS:", pos_tags)
                print("WORD WEIGHTS:", word_weights)
                print("CONSIDERED TOKENS:", considered_tokens, "\n")
                #print("WEIGHTS:", weights, "\n")

        return item_weights, item_considered_tokens
        
    
    def probe_hidden_states(self, verbose:bool = False) -> List[Dict[str, str]]:
        
        # Extract the hidden states
        hidden_states = [item['hs'] for item in self.outputs]
        
        # Weigh the input tokens
        token_weights, considered_tokens = self._weigh_input_tokens(hidden_states, verbose = verbose)
        
        # Initialize the patcher class
        patcher = Patcher(model = self.model, tokenizer = self.tokenizer, prompt = self.target_prompt, placeholder_token = self.placeholder_token)

        # For each input text
        outputs = list()
        
        # TODO CAN BE PARALLELIZED
        for i in tqdm(range(len(self.inputs))):
  
            # Merge the latent representation of the tokens in each layer: [num_layers, num_tokens, hidden_dim]) --> [num_layers, hidden_dim]
            merged_hs = patcher.merged_hs(hidden_states[i], token_weights[i].reshape(-1, 1), verbose = verbose)
            
            # Perform the patching operation for each hidden layer
            generated_texts = patcher.probing(merged_hs)
            
            # Save the output by appending the generated text from the inference and the original input text
            outputs.append({
                'STATS': {'INPUT': self.inputs[i], 'CONSIDERED_TOKENS': considered_tokens[i]},
                '_INFERENCE': self.outputs[i]['text'],
                'HS': generated_texts,
            })
            
        return outputs
    
    def save_stats(self, folder_path:str):
        
        stats = {
            "num_inputs": len(self.inputs),
            "model": self.model_name,
            "num_layers": self.model.config.n_layer,
            "device": self.fabric.device.type.upper() + f"(x{self.num_devices})",
            "prompts": {'source': self.prompts[0].replace(self.inputs[0], "$INPUT"), 'target': self.target_prompt},
            #"model_params": self.model.__dict__
        }
        
        # Create the folder if it does not exist
        if not path.exists(folder_path):
            makedirs(folder_path)
        
        with open(folder_path + "/stats.json", mode = "w", encoding='utf-8') as f:
            dump(stats, f, indent = 4, ensure_ascii=False)


# Copyright Lightning AI. Licensed under the Apache License 2.0, see LICENSE file.
# SOURCE: litgpt/tokenizer.py || VERSION: 0.4.8 || DATA: 2024-08-07

import json
from pathlib import Path
from typing import Optional, Union, List, Dict, Any
import torch

from functools import lru_cache
from litgpt import Tokenizer as litTokenizer

class Tokenizer(litTokenizer):
    
    def __init__(self, checkpoint_dir: Union[Path, str]) -> None:

        # Check if the checkpoint directory exists
        checkpoint_dir = Path(checkpoint_dir)
        if not checkpoint_dir.exists():
            raise NotADirectoryError(f"The checkpoint directory does not exist: {str(checkpoint_dir)}")

        self.model_name = checkpoint_dir.stem

        # Initialize the tokenizer
        self.use_bos = self.check_if_bos_token_used(checkpoint_dir)
        self.bos_id = None
        self.eos_id = None
        
        # Special token path
        if (special_tokens_path := checkpoint_dir / "tokenizer_config.json").is_file():
            with open(special_tokens_path) as fp:
                token_config = json.load(fp)
        
            # Load the special tokens
            self.bos_token = token_config.get("bos_token")
            self.eos_token = token_config.get("eos_token")
            self.additional_special_tokens = token_config.get("additional_special_tokens")
            self.tokenizer_class = token_config.get("tokenizer_class")
            self.chat_template = token_config.get("chat_template")

        # Load the vocabylary and the bos and eos tokens (some checkpoints have both files, `.model` takes precedence)
        
        # TODO ISSUE: The SentencePiece tokenizer does not recognize the special tokens in the prompt chat template!
        # if (vocabulary_path := checkpoint_dir / "tokenizer.model").is_file():
        #    from sentencepiece import SentencePieceProcessor

        #    self.processor = SentencePieceProcessor(model_file=str(vocabulary_path)) #  add_bos=False, add_eos=False
        #    self.backend = "sentencepiece"
            
        #    self.bos_id = self.processor.bos_id()
        #    self.eos_id = self.processor.eos_id()

        if (vocabulary_path := checkpoint_dir / "tokenizer.json").is_file():
            from tokenizers import Tokenizer as HFTokenizer

            self.processor = HFTokenizer.from_file(str(vocabulary_path))
            self.backend = "huggingface"

            if (special_tokens_path := checkpoint_dir / "tokenizer_config.json").is_file():
                
                # Load the special tokens ids (beginning and end of sentence)
                if self.bos_token is not None:
                    if isinstance(self.bos_token, dict):
                        self.bos_id = self.token_to_id(self.bos_token['content'])
                    elif isinstance(self.bos_token, str):
                        self.bos_id = self.token_to_id(self.bos_token)
                        
                if self.eos_token is not None:
                    if isinstance(self.eos_token, dict):
                        self.eos_id = self.token_to_id(self.eos_token['content'])
                    elif isinstance(self.eos_token, str):
                        self.eos_id = self.token_to_id(self.eos_token)
           
        
            #if (special_tokens_path := checkpoint_dir / "special_tokens_map.json").is_file():
            #    with open(special_tokens_path) as fp:
            #        self.special_tokens_map = json.load(fp)
            #else:
            #    self.special_tokens_map = dict()

            if (special_tokens_path := checkpoint_dir / "generation_config.json").is_file():
                with open(special_tokens_path) as fp:
                    gen_config = json.load(fp)
                    
                if self.bos_id is None:
                    self.bos_id = gen_config.get("bos_token_id")
                if self.eos_id is None:
                    self.eos_id = gen_config.get("eos_token_id")
        else:
            raise NotImplementedError('The tokenizer file ("tokenizer_config.json") is not found in the checkpoint directory.')
    
    def encode(
        self,
        string: str,
        device: Optional[torch.device] = None,
        add_special_tokens: bool = True,
    ) -> torch.Tensor:
        if self.backend == "huggingface":
            tokens = self.processor.encode(string, add_special_tokens = add_special_tokens).ids
        elif self.backend == "sentencepiece":
            tokens = self.processor.encode(string)
        else:
            raise RuntimeError(f"`{self.backend}` is not supported.")
        
        if tokens is None:
            raise ValueError("`self.processor` returned tokens of None value.")

        return torch.tensor(tokens, dtype=torch.int, device=device)
    
    
    def id_to_token(self, id: int) -> str:
        if self.backend == "huggingface":
            token = self.processor.id_to_token(id)
        elif self.backend == "sentencepiece":
            token = self.processor.IdToPiece(id)
        else:
            raise RuntimeError
        if id is None:
            raise ValueError(f"id {id!r} not found in the collection.")
        return token
    
    # -----------------------------------------------------------------------------
    # The following methods is copied from the `transformers` library, for applying chat templates
    # -----------------------------------------------------------------------------
    def apply_chat_template(
        self,
        conversation: Union[List[Dict[str, str]], "Conversation"],
        chat_template: Optional[str] = None,
        add_generation_prompt: bool = False,
        tokenize: bool = True,
        padding: bool = False,
        truncation: bool = False,
        max_length: Optional[int] = None,
        return_tensors = None,
        return_dict: bool = False,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[str, List[int]]:
        """
        Converts a Conversation object or a list of dictionaries with `"role"` and `"content"` keys to a list of token
        ids. This method is intended for use with chat models, and will read the tokenizer's chat_template attribute to
        determine the format and control tokens to use when converting. When chat_template is None, it will fall back
        to the default_chat_template specified at the class level.

        Args:
            conversation (Union[List[Dict[str, str]], "Conversation"]): A Conversation object or list of dicts
                with "role" and "content" keys, representing the chat history so far.
            chat_template (str, *optional*): A Jinja template to use for this conversion. If
                this is not passed, the model's default chat template will be used instead.
            add_generation_prompt (bool, *optional*): Whether to end the prompt with the token(s) that indicate
                the start of an assistant message. This is useful when you want to generate a response from the model.
                Note that this argument will be passed to the chat template, and so it must be supported in the
                template for this argument to have any effect.
            tokenize (`bool`, defaults to `True`):
                Whether to tokenize the output. If `False`, the output will be a string.
            padding (`bool`, defaults to `False`):
                Whether to pad sequences to the maximum length. Has no effect if tokenize is `False`.
            truncation (`bool`, defaults to `False`):
                Whether to truncate sequences at the maximum length. Has no effect if tokenize is `False`.
            max_length (`int`, *optional*):
                Maximum length (in tokens) to use for padding or truncation. Has no effect if tokenize is `False`. If
                not specified, the tokenizer's `max_length` attribute will be used as a default.
            return_tensors (`str` or [`~utils.TensorType`], *optional*):
                If set, will return tensors of a particular framework. Has no effect if tokenize is `False`. Acceptable
                values are:
                - `'tf'`: Return TensorFlow `tf.Tensor` objects.
                - `'pt'`: Return PyTorch `torch.Tensor` objects.
                - `'np'`: Return NumPy `np.ndarray` objects.
                - `'jax'`: Return JAX `jnp.ndarray` objects.
            return_dict (`bool`, *optional*, defaults to `False`):
                Whether to return a dictionary with named outputs. Has no effect if tokenize is `False`.
            tokenizer_kwargs (`Dict[str: Any]`, *optional*): Additional kwargs to pass to the tokenizer.
            **kwargs: Additional kwargs to pass to the template renderer. Will be accessible by the chat template.

        Returns:
            `List[int]`: A list of token ids representing the tokenized chat so far, including control tokens. This
            output is ready to pass to the model, either directly or via methods like `generate()`.
        """

        if hasattr(conversation, "messages"):
            # Indicates it's a Conversation object
            conversation = conversation.messages

        if tokenizer_kwargs is None:
            tokenizer_kwargs = {}

        # First, handle the cases when the model has a dict of multiple templates
        if isinstance(self.chat_template, dict) or (
            self.chat_template is None and isinstance(self.default_chat_template, dict)
        ):
            template_dict = self.chat_template or self.default_chat_template
            if chat_template is not None and chat_template in template_dict:
                # The user can pass the name of a template to the chat template argument instead of an entire template
                chat_template = template_dict[chat_template]
            elif chat_template is None and "default" in template_dict:
                chat_template = template_dict["default"]
            elif chat_template is None:
                raise ValueError(
                    "This model has multiple chat templates with no default specified! Please either pass a chat "
                    "template or the name of the template you wish to use to the `chat_template` argument. Available "
                    f"template names are {sorted(template_dict.keys())}."
                )
        elif chat_template is None:
            # These are the cases when the model has a single template
            # priority: `chat_template` argument > `tokenizer.chat_template` > `tokenizer.default_chat_template
            if self.chat_template is not None:
                chat_template = self.chat_template
            else:
                chat_template = self.default_chat_template

        # Compilation function uses a cache to avoid recompiling the same template
        compiled_template = self._compile_jinja_template(chat_template)

        template_kwargs = {**self.special_tokens_map, **kwargs}  # kwargs overwrite special tokens if both are present
        rendered = compiled_template.render(
            messages=conversation, add_generation_prompt=add_generation_prompt, **template_kwargs
        )

        if padding is True:
            padding = "max_length"  # There's only one sequence here, so "longest" makes no sense
        if tokenize:
            if return_dict:
                return self(
                    rendered,
                    padding=padding,
                    truncation=truncation,
                    max_length=max_length,
                    add_special_tokens=False,
                    return_tensors=return_tensors,
                    **tokenizer_kwargs,
                )
            else:
                return self.encode(
                    rendered,
                    #padding=padding,
                    #truncation=truncation,
                    #max_length=max_length,
                    bos = False,
                    eos = False,
                    #add_special_tokens=False,
                    #return_tensors=return_tensors,
                    #**tokenizer_kwargs,
                )
        else:
            return rendered
        
    @lru_cache
    def _compile_jinja_template(self, chat_template):
        try:
            import jinja2
            from jinja2.exceptions import TemplateError
            from jinja2.sandbox import ImmutableSandboxedEnvironment
        except ImportError:
            raise ImportError("apply_chat_template requires jinja2 to be installed.")

        #if version.parse(jinja2.__version__) < version.parse("3.0.0"):
            #raise ImportError("apply_chat_template requires jinja2>=3.0.0 to be installed. Your version is " f"{jinja2.__version__}.")

        def raise_exception(message):
            raise TemplateError(message)

        jinja_env = ImmutableSandboxedEnvironment(trim_blocks=True, lstrip_blocks=True)
        jinja_env.globals["raise_exception"] = raise_exception
        return jinja_env.from_string(chat_template)
    
    @property
    def default_chat_template(self):
        """
        This template formats inputs in the standard ChatML format. See
        https://github.com/openai/openai-python/blob/main/chatml.md
        """

        if self.tokenizer_class in ['GPTNeoXTokenizer']:
            
            # Copied from transformers.models.gpt2.tokenization_gpt2.GPT2Tokenizer.default_chat_template
            return "{% for message in messages %}" "{{ message.content }}{{ eos_token }}" "{% endfor %}"
        else:
            return (
                "{% for message in messages %}"
                "{{'<|im_start|>' + message['role'] + '\n' + message['content'] + '<|im_end|>' + '\n'}}"
                "{% endfor %}"
                "{% if add_generation_prompt %}"
                "{{ '<|im_start|>assistant\n' }}"
                "{% endif %}"
            )

    @property
    def special_tokens_map(self) -> Dict[str, Union[str, List[str]]]:
        """
        `Dict[str, Union[str, List[str]]]`: A dictionary mapping special token class attributes (`cls_token`,
        `unk_token`, etc.) to their values (`'<unk>'`, `'<cls>'`, etc.).

        Convert potential tokens of `tokenizers.AddedToken` type to string.
        """
        
        set_attr = {}
        for attr in ["bos_token", "eos_token", "unk_token", "sep_token", "pad_token", "cls_token", "mask_token", "additional_special_tokens"]:
            if attr in self.__dir__():
                attr_value = getattr(self, attr)

                if attr_value:
                    if isinstance(attr_value, dict) and "content" in attr_value.keys():
                        set_attr[attr] = attr_value['content']
                    elif attr_value is not None:
                        set_attr[attr] = attr_value
        return set_attr


import re
import numpy as np

def _negativeOperator(expressions, verbose = False):
    facts = []
    for negative in expressions:
        negative = negative[:-1] # remove the last parenthesis
        for fact in re.split('∧|∨', negative):
            aux_matches = re.search('is|has', fact, re.IGNORECASE)
            if aux_matches:
                fact = fact[:aux_matches.end()] + 'Not' + fact[aux_matches.end():]
            else:
                fact = 'Not' + fact
            fact = fact.strip(' ¬(')
            facts.append(fact)
            
        if verbose:
            print('NEGATIVE:', facts, " --> PARSED:", facts)
    return facts


def _singleObjectFact(relation, object):
    relationComponents = re.findall('[A-Z][^A-Z]*', relation)

    if object.lower() in ['true', 'false']:
        object2 = object 
        
        if len(relationComponents) > 2:
            obj_pos =  1 if relationComponents[1].lower() not in ['of', 'in'] else relationComponents.index('Of' if relationComponents[1].lower() == 'of' else 'In') + 1
            rel = ' '.join(relationComponents[:obj_pos])
            object1 = ' '.join(relationComponents[obj_pos:])
        else:
            rel = 'is'
            object1 = relation 
    elif len(relationComponents) > 1:
        obj_pos = -1  if relationComponents[0].lower() not in ['is', 'has'] else relationComponents.index('Is' if relationComponents[0].lower() == 'is' else 'Has') + 1
        if 'not' in relationComponents[obj_pos].lower():
            obj_pos += 1

        object1 = object
        rel =  ' '.join(relationComponents[:obj_pos]).replace('_', ' ')
        object2 =  ' '.join(relationComponents[obj_pos:]).replace('_', ' ')
    else:
        object1 = object
        rel = 'is'
        object2 = relationComponents[0]

    return rel, [object1, object2]

def generateSPOTriples(items, verbose = False):

    # isCity(New York City) --> (New York City, is, City) 
    
    # Iterate over the propositions
    triples = []
    for item in items:
        if not isinstance(item, str):
            continue
        
        if verbose:
            print(f"\nRAW:", item)
        
        # Handle the global negation operator ¬() if present 
        global_negation = re.findall(r'¬\(.*\)', item)  # .replace('Not', '¬')
        if len(global_negation) > 0:
            expanded_negative_facts = _negativeOperator(global_negation)
            
            # Remove the negative facts from the original string
            for neg in global_negation:
                item =  item.replace(neg, '')

        # Split the text with the delimiters
        facts = re.split('∧|∨|≠', item)
        
        # Append the unpacked negative facts if present
        if len(global_negation) > 0:
            facts.extend(expanded_negative_facts)
        
        # Parse the facts
        for fact in facts:
            
            # Skip empty facts
            fact = fact.strip()
            if fact == '':
                continue
            
            if verbose:
                print('--> LOGIC:', fact)

            # Extract the parts of the fact
            components = re.split(r'\(', fact)
            
            # Relation
            rel = components[0].strip()

            if len(components) > 2:
                sub_objects = ','.join(components[1:])
                objects = re.split(r',', sub_objects)
            elif len(components) == 2:
                objects = re.split(r',', components[1])
            elif len(components) == 1:
                continue
            else:
                raise NotImplementedError(f'Number of components ({len(components)}) not handled')
            
            objects = [obj.strip(' )') for obj in objects]

            # Clean the strings
            objects = [obj.strip() for obj in objects]
            
            if '¬' in rel:
                rel = rel.replace('¬', '')
                pos = re.search('is|has', rel, re.IGNORECASE)
                if pos:
                    rel = rel[:pos.end()] + 'Not' + rel[pos.end():]
                else:
                    rel = 'Not' + rel
                            
            # Special case with true/false object
            tf_matching = np.argwhere([obj.lower() in ['true', 'false'] for obj in objects]).flatten()
            binary_flag = False
            if len(objects) > 1 and len(tf_matching) > 0:
                obj = objects[tf_matching[0]]
                binary_flag = True
                
                # Modify the relation
                if obj.lower() == 'false':
                    pos = re.search('is|has', rel, re.IGNORECASE)
                    if pos:
                        rel = rel[:pos.end()] + 'Not' + rel[pos.end():]
                    else:
                        rel = 'Not' + rel
                objects.remove(obj)
            
                if len(objects) == 1:
                    rel, objects = _singleObjectFact(rel, objects[0])
                else:
                    rel = rel + objects[0]
                    objects = objects[1:]
                    #raise Exception(f'Unhandled case with {len(objects)} objects ({objects}) --> raw: {fact}')
            
            # Special case
            if len(objects) > 1 and '∧' in objects[0]: 
                objects = objects[0].split('∧') + objects[1:]
            elif len(objects) == 1:
                rel, objects = _singleObjectFact(rel, objects[0])
            
            # Structure the triple for each object
            for obj in objects[1:]:
                
                # Skip the etc
                if 'etc' in obj.lower():
                    continue
  
                # Object order
                if  not binary_flag and ('is' in rel.lower() or 'has' in rel.lower()):
                    triple = (obj, rel, objects[0])
                else:
                    triple = (objects[0], rel, obj)
 
                # Save the SPO triple
                triples.append(triple)
        
                if verbose:
                    print('\tSPO:', triple[0], '<-->', triple[1], '<-->', triple[2])
    return triples


# Copyright Lightning AI. Licensed under the Apache License 2.0, see LICENSE file.
# SOURCE: litgpt/model.py || VERSION: 0.4.8 || DATA: 2024-08-07

from typing import Optional
from torch import Tensor, tanh

from litgpt.config import Config
from litgpt import GPT as litGPT

class GPT(litGPT):
    def __init__(self, config: Config) -> None:
        super().__init__(config)
        
        # New code: initialize the patching operation
        self.init_hidden_states()
        
    # ------------------ NEW CODE [PATCHING] ------------------
    def load_source_hidden_states(self, hidden_states:Tensor = None, patching_token_pos:int = None):
        self.source_hidden_states = hidden_states
        self.patching_token_pos = patching_token_pos
    
    def init_hidden_states(self):
        self.load_source_hidden_states(hidden_states = None, patching_token_pos = None)
        
    def _inject_source_hidden_states(self, hidden_states: Tensor) -> Tensor:
        hidden_states[0, self.patching_token_pos, :] = self.source_hidden_states
        return hidden_states
    # ------------------ NEW CODE [PATCHING] ------------------

    def forward(self, idx: Tensor, input_pos: Optional[Tensor] = None, output_hidden_states:bool = False) -> Tensor:
        T = idx.size(1)
        if self.max_seq_length < T:
            raise ValueError(f"Cannot forward sequence of length {T}, max seq length is only {self.max_seq_length}.")

        # ------------------ NEW CODE ------------------
        all_hidden_states = () if output_hidden_states else None
        # ------------------ NEW CODE ------------------
        
        if input_pos is not None:  # use the kv cache
            cos = self.cos.index_select(0, input_pos)
            sin = self.sin.index_select(0, input_pos)
            if self.mask_cache is None:
                raise TypeError("You need to call `gpt.set_kv_cache()`")
            mask = self.mask_cache.index_select(2, input_pos)
        else:
            cos = self.cos[:T]
            sin = self.sin[:T]
            mask = None

        x = self.transformer.wte(idx)  # token embeddings of shape (b, t, n_embd)
        if self.config.scale_embeddings:
            x = x * Tensor(self.config.n_embd**0.5, dtype=x.dtype)
            
        # ------------------ NEW CODE [PATCHING] ------------------
        first_pass = True if len(input_pos) > 1 else False 
        if first_pass and self.source_hidden_states is not None:
            x = self._inject_source_hidden_states(x)
        # ------------------ NEW CODE [PATCHING] ------------------

        for block in self.transformer.h:
            
            # ------------------ NEW CODE ------------------
            if output_hidden_states:
                all_hidden_states += (x,)
            # ------------------ NEW CODE ------------------
                
            x = block(x, cos, sin, mask, input_pos)
            
        # ------------------ NEW CODE ------------------
        if output_hidden_states:
            all_hidden_states += (x,)
        # ------------------ NEW CODE ------------------
            
        x = self.transformer.ln_f(x)
        x = self.lm_head(x)  # (b, t, vocab_size)

        if self.config.final_logit_softcapping is not None:
            x = tanh(x / self.config.final_logit_softcapping) * self.config.final_logit_softcapping

        if output_hidden_states:
            return x, all_hidden_states
        else:
            return x



import re
import pandas as pd
import matplotlib.pyplot as plt
import networkx as nx
import networkx_temporal as tx
from numpy import ceil
from matplotlib.patches import FancyBboxPatch
    
def clearString(s):
    if not isinstance(s, str) or s == '':
        return s
    
    s = s.replace('_', ' ')
    s = s[0].upper() + s[1:]
    
    if len(re.findall('[A-Z][^A-Z]*', s)) > 0:
        s = ' '.join([word.capitalize().strip() for word in re.findall('[A-Z][^A-Z]*', s)])

    return s
    
def draw_temporal_graph(TG, suptitle = None):
    
    # Get the nodes for the inference layer
    inferece_nodes = TG[-1].nodes()
    
    # Compute the number of rows and columns
    ncols = 5 if len(TG) > 20 else 4
    nrows = ceil(len(TG) / ncols).astype(int)
    spare_suplots = (ncols * nrows) - len(TG)
    
    # Create the figure
    fig, ax = plt.subplots(ncols = ncols, nrows = nrows, figsize=(30, 5 * nrows)) # , constrained_layout=True  

    i, j = 0, 0
    for t, G in enumerate(TG):
        
        # Get the axis
        ax_ = ax[t] if nrows == 1 else ax[i, j]
        ax_.margins(0.1)  
        last = t + 1 == len(TG)

        # Draw the graph
        if G.number_of_nodes() < 8:
            pos = nx.shell_layout(G, scale = 1)
        else:
            pos = nx.circular_layout(G, scale = 1)

        # Draw the graph
        colors = {node: 'firebrick' if node in inferece_nodes else 'indianred' for node in G.nodes()}
        node_labels = {i: '\n'.join(clearString(att['feature']).upper().split()) 
                       if len(att['feature'].split()) < 5 else ' '.join(word.upper() if idx % 2 else f'{word.upper()}\n' 
                                                                     for idx, word in enumerate(clearString(att['feature']).split())) 
                  for i, att in G.nodes(data = True)}
        nx.draw(G, 
                pos = pos, 
                ax = ax_, 
                labels = node_labels,
                node_size = 5000 if G.number_of_nodes() < 8 else 3500, 
                with_labels = True, 
                alpha = 0.9,
                font_size = 10 if G.number_of_nodes() < 8 else 9, 
                edgecolors = list(colors.values()), 
                font_color = 'white', 
                node_color = list(colors.values()),
                edge_color = 'black', 
                arrowsize = 20, 
                width = 2)

        # Draw the edge labels
        edge_lables = {edge: ' '.join(word if idx % 2 else f'{word}\n' 
                                      for idx, word in enumerate(clearString(label).split())).strip() 
                       if len(clearString(label).split()) > 2 else label 
                       for edge, label in nx.get_edge_attributes(G,'label').items()}
        
        nx.draw_networkx_edge_labels(G, pos, ax = ax_, edge_labels = edge_lables, 
                                     font_size = 12 if G.number_of_nodes() < 8 else 10, 
                                     alpha = 0.7 if last else 1, font_color = 'black', bbox = dict(facecolor='white', edgecolor='black', boxstyle='round,pad=0.2'))  
        
        # Add a border around the subplot
        p = FancyBboxPatch((.02, .02), width=.96, height=.96, boxstyle="round,pad=0.02", 
                       fill=False, edgecolor ='black', ls = 'dashed' if last else 'solid', alpha = 0.7 if last else 1,
                       linewidth=1 if last else 2, transform=ax_.transAxes, clip_on=False, zorder = 10)
        ax_.add_patch(p)
        
        # Set the title
        ax_.set_title(label = G.name if 'inference' not in G.name.lower() else 'Inference',
                      fontsize = 16,  color = 'black' if last else 'firebrick', alpha = 0.7 if last else 1)

        # Update the indices
        j += 1
        if ncols and j % ncols == 0:
            j = 0
            i += 1

    # Remove the spare subplots
    for ax in ax.flatten()[len(TG):len(TG) + spare_suplots]:
        ax.remove()
    
    # Add the suptitle
    if suptitle:
        fig.suptitle(suptitle, fontsize=28, fontweight='bold', color = 'firebrick', y = 1)
    
    # Adjust the layout
    fig.tight_layout(h_pad=2)
    
    # Close the figure
    plt.close(fig)
        
    return fig


def generate_temporal_graphs(df):
    
    # Iterate over the inputs
    temporal_graphs = list()
    for _, df_row in df.iterrows():
        
        # Crate the edge list
        edge_df = pd.DataFrame(df_row['hs'].items(), columns = ['layer', 'triple'])
        
        # Add the inference triples
        edge_df.loc[len(edge_df.index)] = ['INFERENCE', df_row['inference']] 
        
        # Unpack the triples
        edge_df = edge_df.explode('triple').reset_index(drop = True)
        unpacked = edge_df['triple'].apply(lambda triple: pd.Series(triple, index = ['source', 'relation', 'target']))
        edge_df = pd.concat([edge_df, unpacked], axis = 1).drop(columns = ['triple'])
        edge_df = edge_df.dropna()
        
        if edge_df.empty:
            continue
        
        # Prepare for the graph
        edge_df['relation'] = edge_df['relation'].map(lambda item: {'label': item})
        
        # Create the graph
        TG = tx.TemporalGraph(directed=True, multigraph=False, t = len(edge_df['layer'].unique()))
        TG.name = df_row['input']
        
        # Add the edges for each layer
        for idk, layer in enumerate(edge_df['layer'].unique()):
            layer_triples = edge_df[edge_df['layer'] == layer].copy()

            # Generate the node ids
            unique_nodes = set(layer_triples['source'].values).union(set(layer_triples['target'].values))
            node_mapping = {nodeName:idk  for idk, nodeName in enumerate(unique_nodes)}
            layer_triples['source'] = layer_triples['source'].map(lambda nodeName: node_mapping[nodeName])
            layer_triples['target'] = layer_triples['target'].map(lambda nodeName: node_mapping[nodeName])

            # Add the nodes and edges
            TG[idk].add_edges_from(layer_triples[['source', 'target', 'relation']].values)
            
            # Add the names to the nodes
            for node_id, attribute in TG[idk].nodes(data = True):
                attribute['feature'] = [nodeName for nodeName, idk in node_mapping.items() if idk == node_id][0]
                
            # Assign the name to the graph 
            TG[idk].name = f"Layer {layer.strip('L')}"
            
        temporal_graphs.append(TG)
    return temporal_graphs


# Copyright Lightning AI. Licensed under the Apache License 2.0, see LICENSE file.
# SOURCE: litgpt/generate/base.py || VERSION: 0.4.8 || DATA: 2024-08-07

from typing import Any, Optional

import torch
import torch._dynamo.config
import torch._inductor.config

from litgpt import GPT
from litgpt.generate.base import sample

@torch.inference_mode()
def generate(
    model: GPT,
    prompt: torch.Tensor,
    max_returned_tokens: int,
    output_hidden_states: bool = False,
    *,
    temperature: float = 1.0,
    top_k: Optional[int] = None,
    top_p: float = 1.0,
    eos_id: Optional[int] = None,
    include_prompt: bool = True,
    **kwargs: Any
) -> torch.Tensor:
    """
    Takes a conditioning sequence (prompt) as input and continues to generate as many tokens as requested.
    The implementation of this function is modified from A. Karpathy's nanoGPT.

    Args:
        model: The model to use.
        prompt: Tensor of shape (T) with indices of the prompt sequence.
        max_returned_tokens: The maximum number of tokens to return (given plus generated).
        temperature: Scales the predicted logits by 1 / temperature.
        top_k: If specified, only sample among the tokens with the k highest probabilities.
        top_p: If specified, it represents the cumulative probability threshold to consider in the sampling process.
            In top-p sampling, the next token is sampled from the highest probability tokens
            whose cumulative probability exceeds the threshold `top_p`. When specified,
            it must be `0 <= top_p <= 1`. Here, `top_p=0` is equivalent
            to sampling the most probable token, while `top_p=1` samples from the whole distribution.
            It can be used in conjunction with `top_k` and `temperature` with the following order
            of application:

            1. `top_k` sampling
            2. `temperature` scaling
            3. `top_p` sampling

            For more details, see https://arxiv.org/abs/1904.09751
            or https://huyenchip.com/2024/01/16/sampling.html#top_p
        eos_id: If specified, stop generating any more token once the <eos> token is triggered.
        include_prompt: If true (default) prepends the prompt (after applying the prompt style) to the output.
    """
    T = prompt.size(0)
    assert max_returned_tokens > T
    if model.max_seq_length < max_returned_tokens - 1:
        # rolling the kv cache based on the `input_pos` value would be necessary. However, doing so would introduce a
        # data dependency on the `input_pos` tensor and impact model compilation. Since this setting is uncommon, we do
        # not support it to avoid negatively impacting the overall speed
        raise NotImplementedError(f"max_seq_length {model.max_seq_length} needs to be >= {max_returned_tokens - 1}")

    device = prompt.device
    tokens = [prompt] if include_prompt else []
    input_pos = torch.tensor([T], device=device)
    
    # ------------------ MODIFIED CODE ---------------------
    # Generate the first token
    token, hidden_states = next_token(
        model = model, 
        input_pos = torch.arange(0, T, device=device), 
        x = prompt.view(1, -1), 
        output_hidden_states = True,
        temperature=temperature, 
        top_k=top_k, 
        # **kwargs     # TODO wait support for other parameters
    )
    token = token.clone()
    tokens.append(token)
    # ------------------ MODIFIED CODE ---------------------
    
    # Generate the rest of the tokens
    for _ in range(2, max_returned_tokens - T + 1):
        token = next_token(model, input_pos, x = token.view(1, -1), temperature=temperature, top_k=top_k).clone() # **kwargs
        tokens.append(token)
        if token == eos_id:
            break
        input_pos = input_pos.add_(1)
    
    # ------------------ MODIFIED CODE --------------------
    if output_hidden_states:
        return torch.cat(tokens), hidden_states
    else:
        return torch.cat(tokens)
    # ------------------ MODIFIED CODE --------------------


def next_token(model: GPT, input_pos: torch.Tensor, x: torch.Tensor, output_hidden_states:bool = False, **kwargs: Any) -> torch.Tensor:
    
    # ------------------ MODIFIED CODE --------------------
    if output_hidden_states:
        logits, hidden_states = model(x, input_pos, output_hidden_states = True)
        hidden_states = [hs.cpu() for hs in hidden_states]
        hidden_states = torch.cat(hidden_states, dim = 0)
    else:
        logits = model(x, input_pos)
    # ------------------ MODIFIED CODE --------------------
    

    next = sample(logits, **kwargs).to(dtype=x.dtype)
    
    # ------------------ MODIFIED CODE --------------------
    if output_hidden_states:
        return next, hidden_states
    else:
        return next
    # ------------------ MODIFIED CODE --------------------




import pandas as pd
from os import path, makedirs
from typing import List, Dict
from networkx_temporal import TemporalGraph

# LOCAL IMPORTS
from .utils.triples import generateSPOTriples
from .utils.graph import generate_temporal_graphs, draw_temporal_graph
    
class TempoGrapher:
    
    def __init__(self, results: List[Dict[str, str]]):
        self.results = results
        self.graphs = self._generate_graphs()

    def _generate_graphs(self) -> List[TemporalGraph]:
        
        # Extract the facts from the results
        df = pd.DataFrame([{
            'input': output['STATS']['INPUT'], 
            'inference': output['_INFERENCE'],
            'hs': output['HS']} 
                            for output in self.results])

        # Generate the SPO triples
        extract_facts = lambda output: output['facts'] if isinstance(output, dict) and 'facts' in output.keys() else list()
        df['inference'] = df['inference'].apply(lambda text: generateSPOTriples(extract_facts(text))) 
        df['hs'] = df['hs'].apply(lambda layer_output: {layer: generateSPOTriples(extract_facts(text)) for layer, text in layer_output.items()}) 
        
        # Perform the knowledge graph generation
        temporal_graphs = generate_temporal_graphs(df = df[['input', 'inference', 'hs']])

        return temporal_graphs
    
    def get_graphs(self) -> List[TemporalGraph]:
        return self.graphs
    
    def save_graphs(self, folder_path:str):
        
        if len(self.graphs) == 0:
            print("No graphs to save!")
            return
        
        # Create the output folder
        folder_path = path.join(folder_path, 'graphs')
        if not path.exists(folder_path):
            makedirs(folder_path)
    
        # Iterate over the claims
        for graph in self.graphs:

            # Create the figure
            fig = draw_temporal_graph(graph, suptitle = graph.name)

            # Create the file name
            cleared_name = graph.name.replace(' ', '_').replace('/', '_').replace('"', '')
            cleared_name = cleared_name[:30] if len(cleared_name) > 30 else cleared_name
            
            # Save the figure
            fig.savefig(path.join(folder_path, f'{cleared_name}.pdf'))
        
        print(f"Graphs saved in {folder_path}")
    